# StepUp RepRep is the StepUp extension for Reproducible Reporting.
# Copyright (C) 2024 Toon Verstraelen
#
# This file is part of StepUp RepRep.
#
# StepUp RepRep is free software;  you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# StepUp RepRep is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Unit tests for stepup.reprep.convert_markdown."""

import pytest

from stepup.reprep.convert_markdown import isolate_header

KATEX = r"""
<ul>
<li>
[1/1]
<math xmlns="http://www.w3.org/1998/Math/MathML">
<semantics>
<mrow><mi>x</mi><mo>=</mo><mfrac>
<mrow><mi>a</mi><msup><mi>t</mi><mn>2</mn></msup></mrow>
<mn>2</mn></mfrac></mrow>
<annotation encoding="application/x-tex">x=\frac{at^2}{2}</annotation>
</semantics>
</math>
</li>
</ul>
""".replace("\n", "")


SPAN_EXAMPLE = '<span style="height:2.476em;vertical-align:-0.9119em;"></span>'


@pytest.mark.parametrize(
    ("orig_body", "header", "body"),
    [
        ("<p>first</p><title>foo</title>", "<title>foo</title>", "<p>first</p>"),
        ("<p>first</p><p>second</p>", "", "<p>first</p><p>second</p>"),
        (
            '<title>foo</title><link rel="stylesheet" href="foo.css" />',
            '<title>foo</title><link href="foo.css" rel="stylesheet"/>',
            "",
        ),
        (KATEX, "", KATEX),
        (SPAN_EXAMPLE, "", SPAN_EXAMPLE),
    ],
)
def test_isolate_header(orig_body, header, body):
    assert isolate_header(orig_body) == (header, body)
    # assert False



# StepUp RepRep is the StepUp extension for Reproducible Reporting.
# Copyright (C) 2024 Toon Verstraelen
#
# This file is part of StepUp RepRep.
#
# StepUp RepRep is free software;  you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# StepUp RepRep is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Markdown to HTML conversion."""

import argparse
import re
import sys

import markdown
from bs4 import BeautifulSoup
from path import Path

from stepup.core.api import amend, getenv, translate_back

from .render import render

__all__ = ("convert_markdown", "isolate_header")


def main(argv: list[str] | None = None):
    """Main program."""
    args = parse_args(argv)
    if not args.markdown.endswith(".md"):
        raise ValueError("The markdown file must end with the .md extension.")
    if args.katex_macros is None and args.katex:
        args.katex_macros = getenv("REPREP_KATEX_MACROS", None)
        if args.katex_macros is not None:
            args.katex_macros = translate_back(args.katex_macros)
            amend(inp=args.katex_macros)
    if args.css is None:
        env_css = getenv("REPREP_MARKDOWN_CSS", "")
        if env_css != "":
            args.css = [translate_back(path_css) for path_css in env_css.split(":")]
    with open(args.markdown) as fm, open(args.html, "w") as fh:
        fh.write(
            convert_markdown(fm.read(), args.katex, args.katex_macros, args.css, args.html.parent)
        )


def parse_args(argv: list[str] | None = None) -> argparse.Namespace:
    """Parse command-line arguments."""
    parser = argparse.ArgumentParser(
        prog="reprep-convert-markdown", description="Convert Markdown to HTML"
    )
    parser.add_argument("markdown", type=Path, help="A Markdown file with extension `.md`")
    parser.add_argument("html", type=Path, help="A HTML output filename")
    parser.add_argument("--katex", default=False, action="store_true", help="Enable KaTeX")
    parser.add_argument(
        "--katex-macros",
        type=Path,
        default=None,
        help="KaTeX macros file. The defualt value if ${REPREP_KATEX_MACROS} (if defined).",
    )
    parser.add_argument(
        "--css",
        type=Path,
        nargs="+",
        default=None,
        help="Local CSS files to link to in the HTML header. "
        "The default value is ${REPREP_MARKDOWN_CSS} (if defined) "
        "and it is interpreted as a colon-separated list",
    )
    return parser.parse_args(argv)


HTML_TEMPLATE = """\
<!DOCTYPE html>
<html>
<head>
  <meta charset="utf-8" />
  <meta name="viewport" content="width=device-width, initial-scale=1.0, user-scalable=yes" />
  <title>{{ title }}</title>
  {{ extra_head | indent(width=2) }}
  {{ css | indent(width=2) }}
</head>
<body>
  {{ body | indent(width=2) }}
</body>
</html>
"""


def convert_markdown(
    text_md: str,
    katex: bool = False,
    path_macro: str | None = None,
    paths_css: str | list[str] | None = None,
    parent_html: str = "./",
) -> str:
    """Convert Markdown to HTML with KaTeX support.

    Parameters
    ----------
    text_md
        The markdown source text.
    katex
        Set to `True` to enable KaTeX support.
    path_macro
        A file with KaTeX macro definitions.
    paths_css
        Path of a local CSS file, or a list of multiple such paths,
        to be included in the HTML header.
        Note that one may also specify CSS file in the markdown header.
    parent_html
        The parent path of the HTML file.
        The CSS paths are rewritten to become relative to this parent path.

    Returns
    -------
    html
        The HTML conversion.
    """
    # Convert conventional LaTeX equation syntax to make it compatible with markdown_katex
    text_md = re.sub(r"\B\$(\S|\S[^\n\r]*?\S)\$\B", r"$`\1`$", text_md)

    extensions = [
        "fenced_code",
        "tables",
        "meta",
    ]
    configs = {}

    if katex:
        extensions.append("markdown_katex")
        configs["markdown_katex"] = {"insert_fonts_css": True}
        if path_macro is not None:
            configs["markdown_katex"]["macro-file"] = path_macro

    md_ctx = markdown.Markdown(
        extensions=extensions,
        extension_configs=configs,
    )

    # Convert to HTML and split into HEAD and BODY parts
    body = md_ctx.convert(text_md)
    extra_head, body = isolate_header(body)

    # Collect all CSS paths
    if paths_css is None:
        paths_css = []
    elif isinstance(paths_css, str):
        paths_css = [paths_css]
    paths_css.extend(md_ctx.Meta.get("css", []))

    # Rewrite CSS paths as paths relative to the parent of the HTML output.
    # parent_html = Path(parent_html)
    paths_css = [
        path_css if "://" in path_css else Path(path_css).relpath(parent_html)
        for path_css in paths_css
    ]

    # Use Jinja to finalize the HTML page.
    variables = {
        "extra_head": extra_head,
        "body": body,
        "title": md_ctx.Meta.get("title", ["Untitled"])[0],
        "css": "\n".join(f'<link rel="stylesheet" href="{path_css}" />' for path_css in paths_css),
    }
    return render("HTML_TEMPLATE", variables, str_in=HTML_TEMPLATE)


def isolate_header(body: str) -> tuple[str, str]:
    """Isolate HTML elements that should be put in the HTML header.

    Normally, the markdown library only produces a HTML body,
    but the markdown_katex plugin needs to insert header elements
    to format the equations correctly.
    This function isolates such header elements,
    so they can be inserted in the right part of the HTML file.

    Parameters
    ----------
    body
        A HTML body with a few header elements.

    Returns
    -------
    header
        A string with HTML header elements.
    body
        The given body without the header elements.
    """
    soup = BeautifulSoup(body, features="html.parser")
    head_lines = []
    body_lines = []
    for child in soup.contents:
        line = str(child).strip()
        if len(line) > 0:
            if child.name in ["link", "style", "meta", "title"]:
                head_lines.append(line)
            else:
                body_lines.append(line)
    return "".join(head_lines), "".join(body_lines)


if __name__ == "__main__":
    main(sys.argv[1:])



#!/usr/bin/env python
from stepup.reprep.api import getenv, latex, render

PUBLIC = getenv("PUBLIC", is_path=True)
render("main.tex", ["../variables.py", "variables.py"], PUBLIC)
latex("main.tex", workdir=PUBLIC)



#!/usr/bin/env python
from stepup.core.api import static
from stepup.reprep.api import latex_flat

static("sub/", "sub/article_structured.tex", "sub/part1.tex", "sub/part2.tex")
latex_flat("sub/article_structured.tex", "sub/article.tex")



#!/usr/bin/env python

from stepup.core.api import static
from stepup.reprep.api import check_hrefs, latex

static("README.md", "main.tex", "check_hrefs.yaml")
latex("main.tex")
check_hrefs("main.pdf")



# StepUp RepRep is the StepUp extension for Reproducible Reporting.
# Copyright (C) 2024 Toon Verstraelen
#
# This file is part of StepUp RepRep.
#
# StepUp RepRep is free software;  you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# StepUp RepRep is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Offline checker for hyper references."""

import argparse
import re
import sys

import attrs
import cattrs.preconf.pyyaml
import fitz
import markdown
import yaml
from bs4 import BeautifulSoup
from path import Path

from stepup.core.api import amend, getenv
from stepup.core.utils import mynormpath


def main(argv: list[str] | None = None):
    """Main program."""
    args = parse_args(argv)
    if args.path_config is None:
        args.path_config = getenv("REPREP_CHECK_HREFS_CONFIG", "check_hrefs.yaml", is_path=True)
        if not amend(inp=args.path_config):
            sys.exit(3)
    config = load_config(args.path_config)
    hrefs = collect_hrefs(args.path_src)
    make_url_substitutions(hrefs, config.subs)
    check_hrefs(hrefs, Path(args.path_src).parent.normpath(), config.accept)


def parse_args(argv: list[str] | None = None) -> argparse.Namespace:
    """Parse command-line arguments."""
    parser = argparse.ArgumentParser(
        prog="reprep-check-hrefs", description="Check hyper references Markdown, HTML or PDF files."
    )
    parser.add_argument("path_src", help="Markdown, HTML or PDF file.")
    parser.add_argument(
        "-c",
        "--config",
        dest="path_config",
        help="Configuration yaml file. "
        "The default is ${REPREP_CHECK_HREFS_CONFIG} or check_hrefs.yaml if it is not set.",
    )
    return parser.parse_args(argv)


@attrs.define
class Config:
    accept: set[str] = attrs.field(factory=set)
    subs: list[tuple[str, str]] = attrs.field(factory=list)


def load_config(path_config) -> Config:
    converter = cattrs.preconf.pyyaml.PyyamlConverter(forbid_extra_keys=True)
    with open(path_config) as fh:
        state = yaml.safe_load(fh)
        return converter.structure(state, Config)


@attrs.define
class HRef:
    """A hyper reference to be checked."""

    url: str = attrs.field()
    translated: str = attrs.field(default=None)
    ignore: bool = attrs.field(default=False, init=False)


def collect_hrefs(path_src: str) -> list[HRef]:
    """Find all hyper references in one file."""
    if path_src.endswith(".md"):
        result = collect_hrefs_md(path_src)
    elif path_src.endswith(".html"):
        result = collect_hrefs_html(path_src)
    elif path_src.endswith(".pdf"):
        result = collect_hrefs_pdf(path_src)
    else:
        raise ValueError(f"Source file type not supported: {path_src}")
    return result


def collect_hrefs_md(fn_md: str) -> list[HRef]:
    """Find all hyper references in a Markdown file."""
    with open(fn_md) as fh:
        html = markdown.markdown(fh.read())
    soup = BeautifulSoup(html, "html.parser")
    return [HRef(link.attrs["href"]) for link in soup.find_all("a")]


def collect_hrefs_html(fn_html: str) -> list[HRef]:
    """Find all hyper references in a HTML file."""
    with open(fn_html) as fh:
        soup = BeautifulSoup(fh.read(), "html.parser")
    return [HRef(link.attrs["href"]) for link in soup.find_all("a")]


def collect_hrefs_pdf(fn_pdf: str) -> list[HRef]:
    """Find all hyper references in a PDF file."""
    # See https://pymupdf.readthedocs.io/en/latest/page.html#description-of-get-links-entries
    doc = fitz.open(fn_pdf)
    hrefs = []
    for page in doc:
        for link in page.get_links():
            if link["kind"] in (fitz.LINK_LAUNCH, fitz.LINK_GOTOR):
                hrefs.append(HRef(link["file"]))
            elif link["kind"] == fitz.LINK_URI:
                hrefs.append(HRef(link["uri"]))
    return hrefs


def make_url_substitutions(hrefs: list[HRef], subs: list[tuple[str, str]]):
    for href in hrefs:
        url = href.url
        if "#" in href.url:
            url = url[: href.url.find("#")]
        for pattern, replacement in subs:
            url = re.sub(pattern, replacement, url)
        href.translated = url


def check_hrefs(hrefs: list[HRef], root: Path, accept: set[str]):
    """Check the hyper references."""
    seen = set()
    some_failed = False
    for href in hrefs:
        if href.url in seen:
            continue
        some_failed |= check_href(href, root, accept)
        seen.add(href.url)
    if some_failed:
        raise ValueError("Some hyper references were invalid.")


def check_href(href: HRef, root: Path, accept: set[str]) -> bool:
    """Check a hyper reference and return True if an error was found."""
    if href.url.startswith("mailto:"):
        return False
    if "://" in href.translated:
        if href.translated in accept:
            return False
        print(f"FAILED LINK: {href.url}", file=sys.stderr)
        return True
    path = mynormpath(root / Path(href.translated))
    amend(inp=path)
    return False


if __name__ == "__main__":
    sys.exit(main())



# StepUp RepRep is the StepUp extension for Reproducible Reporting.
# Copyright (C) 2024 Toon Verstraelen
#
# This file is part of StepUp RepRep.
#
# StepUp RepRep is free software;  you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# StepUp RepRep is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Unit tests for stepup.reprep.sync_zenodo."""

from stepup.reprep.sync_zenodo import ZenodoWrapper


def test_links():
    zenodo = ZenodoWrapper("")
    links = {"foo": f"{zenodo.endpoint}/foo"}
    zenodo._normalize_links(links)
    assert links == {"foo": "foo"}



#!/usr/bin/env python
from stepup.core.api import static
from stepup.reprep.api import convert_markdown

static("sub/", "sub/demo.md", "demo.css", "page.css", "macros.tex")
convert_markdown("sub/demo.md", katex=True, path_macro="macros.tex", paths_css="page.css")



# StepUp RepRep is the StepUp extension for Reproducible Reporting.
# Copyright (C) 2024 Toon Verstraelen
#
# This file is part of StepUp RepRep.
#
# StepUp RepRep is free software;  you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# StepUp RepRep is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Wrapper for HTML to PDF conversion."""

import argparse
import sys
from collections.abc import Iterator

from defusedxml import ElementTree
from path import Path

from stepup.core.api import getenv, step


def main(argv: list[str] | None = None):
    """Main program."""
    args = parse_args(argv)
    if not args.path_pdf.endswith(".pdf"):
        raise ValueError("The output must have a pdf extensions.")
    if args.weasyprint is None:
        args.weasyprint = getenv("REPREP_WEASYPRINT", "weasyprint")
    convert_html_pdf(args.path_html, args.path_pdf, args.weasyprint, args.optional)


def parse_args(argv: list[str] | None = None) -> argparse.Namespace:
    """Parse command-line arguments."""
    parser = argparse.ArgumentParser(
        prog="reprep-convert-weasyprint",
        description="Convert a HTML to PDF, with dependency tracking.",
    )
    parser.add_argument("path_html", help="The input HTML file.")
    parser.add_argument("path_pdf", help="The output PDF file.")
    parser.add_argument(
        "--weasyprint",
        help="The weasyprint executable to use. "
        "Defaults to `${REPREP_WEASYPRINT}` variable or `weasyprint` if the variable is unset.",
    )
    parser.add_argument(
        "--optional",
        default=False,
        action="store_true",
        help="With this option, the conversion becomes optional.",
    )
    return parser.parse_args(argv)


def convert_html_pdf(path_html: str, path_pdf: str, weasyprint: str, optional: bool):
    inp_paths = search_html_deps(path_html)
    step(
        f"{weasyprint} {path_html} {path_pdf}",
        inp=[path_html, *inp_paths],
        out=path_pdf,
        optional=optional,
    )


def search_html_deps(src: str) -> list[str]:
    """Search implicit dependencies in HTML files: included images and css."""
    implicit = []
    todo = [src]
    idep = 0
    while idep < len(todo):
        path_html = Path(todo[idep])
        for href in iter_html_hrefs(path_html):
            if href.startswith("file://"):
                href = href[7:]
            if "://" not in href:
                if not href.startswith("/"):
                    href = path_html.parent / href
                implicit.append(href)
                if href.endswith(".html"):
                    todo.append(href)
        idep += 1
    return implicit


def iter_html_hrefs(path_html: str) -> Iterator[str]:
    parser = ElementTree.iterparse(path_html, events=("start",))
    for event, elem in parser:
        if event == "start":
            tag = elem.tag.rpartition("}")[2]
            if tag == "img":
                for key in "src", "{http://www.w3.org/1999/xhtml}src":
                    href = elem.attrib.get(key)
                    if href is not None and "data:image" not in href:
                        yield href
                        break
            elif tag == "link":
                for key in "href", "{http://www.w3.org/1999/xhtml}href":
                    href = elem.attrib.get(key)
                    if href is not None and "data:image" not in href:
                        yield href
                        break
        elem.clear()


if __name__ == "__main__":
    main(sys.argv[1:])



#!/usr/bin/env python
from stepup.core.api import copy, getenv, glob, mkdir, plan, static

PUBLIC = getenv("PUBLIC", is_path=True)
glob("static/**", _defer=True)
static("variables.py")
mkdir(PUBLIC)
copy("static/preamble.inc.tex", PUBLIC, optional=True)
plan("static/")



# StepUp RepRep is the StepUp extension for Reproducible Reporting.
# Copyright (C) 2024 Toon Verstraelen
#
# This file is part of StepUp RepRep.
#
# StepUp RepRep is free software;  you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# StepUp RepRep is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Wrapper for SVG to PDF conversion."""

import argparse
import sys
from collections.abc import Iterator

from defusedxml import ElementTree
from path import Path

from stepup.core.api import getenv, step


def main(argv: list[str] | None = None):
    """Main program."""
    args = parse_args(argv)
    path_out = Path(args.path_out)
    allowed_extensions = [".pdf", ".png"]
    if not any(path_out.endswith(ext) for ext in allowed_extensions):
        raise ValueError(
            f"The output must have one of the following extensions: {allowed_extensions}"
        )
    if args.inkscape is None:
        args.inkscape = getenv("REPREP_INKSCAPE", "inkscape")
    if len(args.inkscape_args) == 0:
        inkscape_args = getenv(f"REPREP_INKSCAPE_{path_out.suffix[1:].upper()}_ARGS", "")
    else:
        inkscape_args = " ".join(args.inkscape_args)
    convert_svg_pdf(args.path_svg, path_out, args.inkscape, inkscape_args, args.optional)


def parse_args(argv: list[str] | None = None) -> argparse.Namespace:
    """Parse command-line arguments."""
    parser = argparse.ArgumentParser(
        prog="reprep-convert-inkscape",
        description="Convert an SVG to PDF or PNG, with dependency tracking.",
    )
    parser.add_argument("path_svg", help="The input SVG file.")
    parser.add_argument("path_out", help="The output PDF or PNG file.")
    parser.add_argument(
        "inkscape_args",
        nargs="*",
        help="Additional arguments to be passed into inkscape. "
        "E.g. -T (for text to path conversion). "
        "Depending on the output extension, the defaults is `${REPREP_INKSCAPE_PDF_ARGS}` or "
        "`${REPREP_INKSCAPE_PDF_ARGS}`, if the environment variable is defined.",
    )
    parser.add_argument(
        "--inkscape",
        help="The inkscape executable to use. "
        "Defaults to `${REPREP_INKSCAPE}` variable or `inkscape` if the variable is unset.",
    )
    parser.add_argument(
        "--optional",
        default=False,
        action="store_true",
        help="With this option, the conversion becomes optional.",
    )
    return parser.parse_args(argv)


def convert_svg_pdf(
    path_svg: str, path_out: Path, inkscape: str, inkscape_args: str, optional: bool
):
    inp_paths = search_svg_deps(path_svg)
    fmt = path_out.suffix[1:]
    step(
        f"SELF_CALL=x {inkscape} {path_svg} {inkscape_args} "
        f"--export-filename={path_out} --export-type={fmt}",
        inp=[path_svg, *inp_paths],
        out=path_out,
        optional=optional,
    )


def search_svg_deps(src: str) -> list[str]:
    """Search implicit dependencies in SVG files, specifically (recursively) included images."""
    implicit = []
    todo = [src]
    idep = 0
    while idep < len(todo):
        path_svg = Path(todo[idep])
        for href in iter_svg_image_hrefs(path_svg):
            if href.startswith("file://"):
                href = href[7:]
            if "://" not in href:
                if not href.startswith("/"):
                    href = path_svg.parent / href
                implicit.append(href)
                if href.endswith(".svg"):
                    todo.append(href)
        idep += 1
    return implicit


def iter_svg_image_hrefs(path_svg: str) -> Iterator[str]:
    parser = ElementTree.iterparse(path_svg, events=("start",))
    for event, elem in parser:
        if event == "start":
            tag = elem.tag.rpartition("}")[2]
            if tag == "image":
                for key in "href", "{http://www.w3.org/1999/xlink}href":
                    href = elem.attrib.get(key)
                    if href is not None and "data:image" not in href:
                        yield href
                        break
        elem.clear()


if __name__ == "__main__":
    main(sys.argv[1:])



#!/usr/bin/env python
from stepup.core.api import static, step
from stepup.reprep.api import make_inventory, zip_inventory

static("static.txt")
step("echo hello > ${out}", out="built.txt")
make_inventory(["static.txt", "built.txt"], "inventory.txt")
zip_inventory("inventory.txt", "upload.zip")



#!/usr/bin/env python

from stepup.core.api import mkdir, static, step
from stepup.reprep.api import latex

static("paper.tex")
mkdir("subdir/")
latex("paper.tex")
step(
    "echo 'Verbatim input:\\verbatiminput{code.txt}' > subdir/generated.tex",
    out=["subdir/generated.tex"],
)
step("echo '2 + 2' > subdir/code.txt", out=["subdir/code.txt"])



# StepUp RepRep is the StepUp extension for Reproducible Reporting.
# Copyright (C) 2024 Toon Verstraelen
#
# This file is part of StepUp RepRep.
#
# StepUp RepRep is free software;  you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# StepUp RepRep is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Application programming interface for StepUp RepRep."""

from collections.abc import Collection

from stepup.core.api import StepInfo, getenv, step, subs_env_vars
from stepup.core.utils import make_path_out

__all__ = (
    "cat_pdf",
    "check_hrefs",
    "convert_markdown",
    "convert_weasyprint",
    "convert_odf_pdf",
    "convert_pdf",
    "convert_pdf_png",
    "convert_svg",
    "convert_svg_pdf",
    "convert_svg_png",
    "latex",
    "latex_diff",
    "latex_flat",
    "make_inventory",
    "nup_pdf",
    "raster_pdf",
    "render",
    "sync_zenodo",
    "unplot",
    "zip_inventory",
)


def add_notes_pdf(
    path_src: str, path_notes: str, path_dst: str, optional: bool = False, block: bool = False
) -> StepInfo:
    """Add a notes page at every even page of a PDF file.

    Parameters
    ----------
    path_src
        The original PDF document without notes pages.
    path_notes
        A single-page PDF document with a page suitable for taking notes.
    path_dst
        The output PDF with notes pages inserted.
    optional
        When `True`, the step is only executed when needed by other steps.
    block
        When `True`, the step will always remain pending.

    Returns
    -------
    step_info
        Holds relevant information of the step, useful for defining follow-up steps.
    """
    return step(
        "python -m stepup.reprep.add_notes_pdf ${inp} ${out}",
        inp=[path_src, path_notes],
        out=path_dst,
        optional=optional,
        block=block,
    )


def cat_pdf(
    paths_inp: Collection[str],
    path_out: str,
    *,
    mutool: str | None = None,
    optional: bool = False,
    block: bool = False,
) -> StepInfo:
    """Concatenate the pages of multiple PDFs into one document

    Parameters
    ----------
    paths_inp
        The input PDF files.
    path_out
        The concatenated PDF.
    mutool
        The path to the mutool executable.
        Defaults to `${REPREP_MUTOOL}` variable or `mutool` if the variable is unset.
    optional
        When `True`, the step is only executed when needed by other steps.
    block
        When `True`, the step will always remain pending.

    Returns
    -------
    step_info
        Holds relevant information of the step, useful for defining follow-up steps.
    """
    if mutool is None:
        mutool = getenv("REPREP_MUTOOL", "mutool")
    return step(
        mutool + " merge -o ${out} ${inp}",
        inp=paths_inp,
        out=path_out,
        optional=optional,
        block=block,
    )


def check_hrefs(path_src: str, path_config: str | None = None, block: bool = False) -> StepInfo:
    """Check hyper references in a Markdown, HTML or PDF file.

    Parameters
    ----------
    path_src
        The source Markdown, HTML or PDF to check.
    path_config
        The configuration file.
        Defaults to `${REPREP_CHECK_HREFS_CONFIG}` variable or `check_hrefs.yaml` if it is not set.
    block
        When `True`, the step will always remain pending.

    Returns
    -------
    step_info
        Holds relevant information of the step, useful for defining follow-up steps.
    """
    with subs_env_vars() as subs:
        path_src = subs(path_src)
        path_config = subs(path_config)
    command = f"python -m stepup.reprep.check_hrefs {path_src}"
    inp_paths = [path_src]
    if path_config is not None:
        inp_paths.append(path_config)
        command += f" -c {path_config}"
    return step(command, inp=inp_paths, block=block)


def convert_markdown(
    path_md: str,
    out: str | None = None,
    *,
    katex: bool = False,
    path_macro: str | None = None,
    paths_css: str | list[str] | None = None,
    optional: bool = False,
    block: bool = False,
) -> StepInfo:
    """Convert a markdown to HTML.

    Parameters
    ----------
    path_md
        The markdown input file.
    out
        Output destination: `None`, a directory or a file.
    katex
        Set to `True` to enable KaTeX support.
    path_macro
        A file with macro definitions for KaTeX.
        Defaults to `${REPREP_KATEX_MACROS}` if the variable is set.
    paths_css
        Path of a local CSS file, or a list of multiple such paths,
        to be included in the HTML header.
        Note that one may also specify CSS file in the markdown header.
        Defaults to `${REPREP_MARKDOWN_CSS}` if the variable is set,
        which is interpreted as a colon-separated list of files.
    optional
        When `True`, the step is only executed when needed by other steps.
    block
        When `True`, the step will always remain pending.

    Returns
    -------
    step_info
        Holds relevant information of the step, useful for defining follow-up steps.
    """
    with subs_env_vars() as subs:
        path_md = subs(path_md)
        out = subs(out)
    if not path_md.endswith(".md"):
        raise ValueError("The Markdown file must have extension .md")
    path_html = make_path_out(path_md, out, ".html")
    inp = [path_md]
    command = f"python -m stepup.reprep.convert_markdown {path_md} {path_html}"
    if katex:
        command += " --katex"
        if path_macro is not None:
            command += f" --katex-macros={path_macro}"
            inp.append(path_macro)
    if paths_css is not None:
        if isinstance(paths_css, str):
            paths_css = [paths_css]
        command += " --css " + " ".join(paths_css)
    return step(command, inp=inp, out=path_html, optional=optional, block=block)


def convert_weasyprint(
    path_html: str,
    out: str | None = None,
    *,
    weasyprint: str | None = None,
    optional: bool = False,
    block: bool = False,
) -> StepInfo:
    """Convert a HTML document to PDF.

    Parameters
    ----------
    path_html
        The HTML input file.
    out
        Output destination: `None`, a directory or a file.
    weasyprint
        The path to the weasyprint executable.
        Defaults to `${REPREP_WEASYPRINT}` variable or `weasyprint` if the variable is unset.
    optional
        When `True`, the step is only executed when needed by other steps.
    block
        When `True`, the step will always remain pending.

    Returns
    -------
    step_info
        Holds relevant information of the step, useful for defining follow-up steps.
    """
    with subs_env_vars() as subs:
        path_html = subs(path_html)
        out = subs(out)
    if not path_html.endswith(".html"):
        raise ValueError("The HTML file must have extension .html")
    path_pdf = make_path_out(path_html, out, ".pdf")
    command = f"python -m stepup.reprep.convert_weasyprint {path_html} {path_pdf}"
    if weasyprint is not None:
        command += f" --weasyprint={weasyprint}"
    if optional:
        command += " --optional"
    return step(command, inp=path_html, block=block)


def convert_odf_pdf(
    path_odf: str,
    out: str | None = None,
    *,
    libreoffice: str | None = None,
    optional: bool = False,
    block: bool = False,
) -> StepInfo:
    """Convert a file in OpenDocument format to PDF.

    Parameters
    ----------
    path_odf
        The input open-document file.
    out
        None, output directory or path. See `make_path_out`.
    libreoffice
        The libreoffice executable.
        Defaults to `${REPREP_LIBREOFFICE}` variable or `libreoffice` if the variable is unset.
    optional
        When `True`, the step is only executed when needed by other steps.
    block
        When `True`, the step will always remain pending.

    Returns
    -------
    step_info
        Holds relevant information of the step, useful for defining follow-up steps.

    Notes
    -----
    This function does not yet scan the source document for reference to external files.
    which should ideally be added as dependencies.

    The conversion is executed in a pool of size 1, due to a bug in libreoffice.
    It cannot perform multiple PDF conversions in parallel.
    """
    with subs_env_vars() as subs:
        path_odf = subs(path_odf)
        out = subs(out)
    if libreoffice is None:
        libreoffice = getenv("REPREP_LIBREOFFICE", "libreoffice")
    command = (
        # Simple things should be simple! ;) See:
        # https://bugs.documentfoundation.org/show_bug.cgi?id=106134
        # https://bugs.documentfoundation.org/show_bug.cgi?id=152192
        # Not solved yet:
        # https://bugs.documentfoundation.org/show_bug.cgi?id=160033
        f"WORK=`mktemp -d --suffix=reprep` && {libreoffice} "
        "-env:UserInstallation=file://${WORK} --convert-to pdf ${inp} --outdir ${WORK} "
        "> /dev/null && cp ${WORK}/*.pdf ${out} && rm -r ${WORK}"
    )
    path_pdf = make_path_out(path_odf, out, ".pdf")
    return step(command, inp=path_odf, out=path_pdf, optional=optional, block=block)


def convert_pdf(
    path_pdf: str,
    path_out: str,
    *,
    resolution: int | None = None,
    mutool: str | None = None,
    optional: bool = False,
    block: bool = False,
) -> StepInfo:
    """Convert a PDF to a bitmap with mutool (from MuPDF).

    Parameters
    ----------
    path_pdf
        The input PDF file.
    path_out
        The output image file.
    resolution
        The resolution of the output bitmap in dots per inch (dpi).
    mutool
        The path to the mutool executable.
        Defaults to `${REPREP_MUTOOL}` variable or `mutool` if the variable is unset.
    optional
        When `True`, the step is only executed when needed by other steps.
    block
        When `True`, the step will always remain pending.

    Returns
    -------
    step_info
        Holds relevant information of the step, useful for defining follow-up steps.
    """
    if resolution is None:
        resolution = int(getenv("REPREP_CONVERT_PDF_RESOLUTION", "100"))
    if mutool is None:
        mutool = getenv("REPREP_MUTOOL", "mutool")
    return step(
        f"{mutool} draw -q -o ${{out}} -r {resolution} ${{inp}}",
        inp=path_pdf,
        out=path_out,
        optional=optional,
        block=block,
    )


def convert_pdf_png(
    path_pdf: str,
    out: str | None = None,
    *,
    resolution: int | None = None,
    mutool: str | None = None,
    optional: bool = False,
    block: bool = False,
) -> StepInfo:
    """Shorthand for `convert_pdf` with the output file derived from the PDF file.

    The `out` argument can be `None`, a directory or a file. See `make_path_out`.
    """
    with subs_env_vars() as subs:
        path_pdf = subs(path_pdf)
        out = subs(out)
    if not path_pdf.endswith(".pdf"):
        raise ValueError("The PDF file must have extension .pdf")
    path_png = make_path_out(path_pdf, out, ".png")
    return convert_pdf(
        path_pdf, path_png, resolution=resolution, mutool=mutool, optional=optional, block=block
    )


def convert_svg(
    path_svg: str,
    path_out: str,
    *,
    inkscape: str | None = None,
    inkscape_args: str | None = None,
    optional: bool = False,
    block: bool = False,
) -> StepInfo:
    """Convert an SVG figure to a PDF file, detecting dependencies of the SVG on other files.

    Parameters
    ----------
    path_svg
        The input SVG figure. It may contain <img> tags referring to other files included in
        the figure.
    path_out
        The output PDF or PNG file. Other formats are not supported.
    inkscape
        The path to the inkscape executable.
        Defaults to `${REPREP_INKSCAPE}` variable or `inkscape` if the variable is unset.
    inkscape_args
        Additional arguments to pass to inkscape. E.g. `-T` to convert text to glyphs in PDFs.
        Depending on the extension of the output, the default is `${REPREP_INKSCAPE_PDF_ARGS}` or
        `${REPREP_INKSCAPE_PNG_ARGS}`, if the environment variable is defined.
    optional
        When `True`, the step is only executed when needed by other steps.
    block
        When `True`, the step will always remain pending.

    Returns
    -------
    step_info
        Holds relevant information of the step, useful for defining follow-up steps.

    Notes
    -----
    A wrapper around inkscape is used to carry out the conversion: `stepup.reprep.convert_svg_pdf`.
    The wrapper scans the SVG for dependencies, which may be a bit slow in case of large files.
    """
    with subs_env_vars() as subs:
        path_svg = subs(path_svg)
        path_out = subs(path_out)
    if not path_svg.endswith(".svg"):
        raise ValueError("The SVG file must have extension .svg")
    if not path_out.endswith((".pdf", ".png")):
        raise ValueError("The output file must have extension .pdf or .png")
    command = f"python -m stepup.reprep.convert_inkscape {path_svg} {path_out}"
    if inkscape is not None:
        command += f" --inkscape={inkscape}"
    if inkscape_args is not None:
        command += f" -- {inkscape_args}"
    if optional:
        command += " --optional"
    return step(command, inp=path_svg, block=block)


def convert_svg_pdf(
    path_svg: str,
    out: str | None = None,
    *,
    inkscape: str | None = None,
    inkscape_args: str | None = None,
    optional: bool = False,
    block: bool = False,
) -> StepInfo:
    """Shorthand for `convert_svg` with the output file derived from the SVG file.

    The `out` argument can be `None`, a directory or a file.
    """
    with subs_env_vars() as subs:
        path_svg = subs(path_svg)
        out = subs(out)
    if not path_svg.endswith(".svg"):
        raise ValueError("The SVG file must have extension .svg")
    path_pdf = make_path_out(path_svg, out, ".pdf")
    return convert_svg(
        path_svg,
        path_pdf,
        inkscape=inkscape,
        inkscape_args=inkscape_args,
        optional=optional,
        block=block,
    )


def convert_svg_png(
    path_svg: str,
    out: str | None = None,
    *,
    inkscape: str | None = None,
    inkscape_args: str | None = None,
    optional: bool = False,
    block: bool = False,
) -> StepInfo:
    """Shorthand for `convert_svg` with the output file derived from the SVG file.

    The `out` argument can be `None`, a directory or a file. See `make_path_out`.
    """
    with subs_env_vars() as subs:
        path_svg = subs(path_svg)
        out = subs(out)
    if not path_svg.endswith(".svg"):
        raise ValueError("The SVG file must have extension .svg")
    path_png = make_path_out(path_svg, out, ".png")
    return convert_svg(
        path_svg,
        path_png,
        inkscape=inkscape,
        inkscape_args=inkscape_args,
        optional=optional,
        block=block,
    )


def latex(
    path_tex: str,
    *,
    run_bibtex=True,
    maxrep: int = 5,
    workdir: str = "./",
    latex: str | None = None,
    bibtex: str | None = None,
    bibsane: str | None = None,
    bibsane_config: str | None = None,
    optional: bool = False,
    block: bool = False,
) -> StepInfo:
    """Create a step for the compilation of a LaTeX source.

    Parameters
    ----------
    path_tex
        The main tex source file.
        This argument may contain environment variables.
    run_bibtex
        By default, when bib files are used, BibTeX is invoked.
        This can be overruled by setting this argument to False,
        which is useful when recompiling sources with fixed bbl files.
    maxrep
        The maximum number of repetitions of the LaTeX command
        in case the aux file keeps changing.
    workdir
        The working directory where the LaTeX command must be executed.
    latex
        Path to the LaTeX executable. Note that only PDF-producing LaTeX compilers are supported:
        `pdflatex`, `xelatex` or `lualatex`.
        Defaults to `${REPREP_LATEX}` variable or `pdflatex` if the variable is unset.
    bibtex
        Path to the BibTeX executable.
        Defaults to `${REPREP_BIBTEX}` variable or `bibtex` if the variable is unset.
    bibsane
        Path to the BibSane executable.
        Defaults to `${REPREP_BIBSANE}` variable or `bibsane` if the variable is unset.
    bibsane_config
        Path to the BibSane configuration file.
        Defaults to `${REPREP_BIBSANE_CONFIG}` variable or `bibsane.yaml` if it is unset.
        Note that when the config file is read from the environment variable,
        it is interpreted relative to `${STEPUP_ROOT}`.
        One may define it globally with `export REPREP_BIBSANE_CONFIG='${HERE}/bibsane.yaml'`
        to refer to a local version of the file. (Mind the single quotes.)
    optional
        When `True`, the step is only executed when needed by other steps.
    block
        When `True`, the step will always remain pending.

    Returns
    -------
    step_info
        Holds relevant information of the step, useful for defining follow-up steps.
    """
    with subs_env_vars() as subs:
        path_tex = subs(path_tex)
    if not path_tex.endswith(".tex"):
        raise ValueError(f"The input of the latex command must end with .tex, got {path_tex}.")

    prefix = path_tex[:-4]
    path_pdf = f"{prefix}.pdf"

    command = f"python -m stepup.reprep.latex {path_tex}"
    inp_paths = [path_tex]
    if maxrep != 5:
        command += f" --maxrep={maxrep}"
    if latex is not None:
        command += f" --latex={latex}"
    if run_bibtex:
        command += " --run-bibtex"
        if bibtex is not None:
            command += f" --bibtex={bibtex}"
        if bibsane is not None:
            command += f" --bibsane={bibsane}"
        if bibsane_config is not None:
            command += f" --bibsane-config={bibsane_config}"
            inp_paths.append(bibsane_config)
    return step(
        command,
        inp=inp_paths,
        out=[path_pdf, f"{prefix}.aux", f"{prefix}-inventory.txt"],
        workdir=workdir,
        optional=optional,
        block=block,
    )


DEFAULT_LATEXDIFF_ARGS = (
    "--append-context2cmd=abstract,supplementary,dataavailability,funding,"
    "authorcontributions,conflictsofinterest,abbreviations"
)


def latex_diff(
    path_old: str,
    path_new: str,
    path_diff: str,
    *,
    latexdiff: str | None = None,
    latexdiff_args: str | None = DEFAULT_LATEXDIFF_ARGS,
    optional: bool = False,
    block: bool = False,
) -> StepInfo:
    r"""Create a step to run latexdiff.

    Parameters
    ----------
    path_old
        The old tex or bbl source.
    path_new
        The new tex or bbl source.
    path_diff
        The diff output tex or bbl.
    latexdiff
        Path of the latexdiff  executable.
        Defaults to `${REPREP_LATEXDIFF}` variable or `latexdiff` if the variable is unset.
    latexdiff_args
        Additional arguments for latexdiff.
        Defaults to `${REPREP_LATEXDIFF_ARG}` variable.
        If this variable is unset, the following default is used:

        ```
        --append-context2cmd=abstract,supplementary,dataavailability,funding, \
                             authorcontributions,conflictsofinterest,abbreviations
        ```

        The option `--no-label` is always added because it is needed to make the file reproducible.
    optional
        When `True`, the step is only executed when needed by other steps.
    block
        When `True`, the step will always remain pending.

    Returns
    -------
    step_info
        Holds relevant information of the step, useful for defining follow-up steps.
    """
    if latexdiff is None:
        latexdiff = getenv("REPREP_LATEXDIFF", "latexdiff")

    if latexdiff_args is None:
        latexdiff_args = getenv("REPREP_LATEXDIFF_ARGS", "")

    return step(
        f"{latexdiff} {latexdiff_args} ${{inp}} --no-label > ${{out}}",
        inp=[path_old, path_new],
        out=path_diff,
        optional=optional,
        block=block,
    )


def latex_flat(path_tex: str, path_flat: str, *, optional: bool = False, block: bool = False):
    r"""Flatten structured LaTeX source files (substitute `\input` and friends by their content).

    Parameters
    ----------
    path_tex
        The main tex file to be converted.
    path_flat
        The flattened output file.
    optional
        When `True`, the step is only executed when needed by other steps.
    block
        When `True`, the step will always remain pending.

    Returns
    -------
    step_info
        Holds relevant information of the step, useful for defining follow-up steps.
    """
    return step(
        "python -m stepup.reprep.latex_flat ${inp} ${out}",
        inp=path_tex,
        out=path_flat,
        optional=optional,
        block=block,
    )


def make_inventory(
    paths: Collection[str], path_inventory: str, *, optional: bool = False, block: bool = False
) -> StepInfo:
    """Create an `inventory.txt` file.

    Parameters
    ----------
    paths
        Paths to include in the `inventory.txt` file.
    path_inventory
        The inventory file to write.
    optional
        When `True`, the step is only executed when needed by other steps.
    block
        When `True`, the step will always remain pending.

    Returns
    -------
    step_info
        Holds relevant information of the step, useful for defining follow-up steps.
    """
    return step(
        "reprep-make-inventory ${inp} -o ${out}",
        inp=paths,
        out=[path_inventory],
        optional=optional,
        block=block,
    )


def nup_pdf(
    path_src: str,
    path_dst: str,
    *,
    nrow: int | None = None,
    ncol: int | None = None,
    margin: float | None = None,
    page_format: str | None = None,
    optional: bool = False,
    block: bool = False,
) -> StepInfo:
    """Put multiple pages per sheet using a fixed layout.

    Parameters
    ----------
    path_src
        The original PDF document (with normal pages).
    path_dst
        The output PDF with (multiple pages per sheet).
    nrow
        The number of rows on each output sheet.
        The default is `${REPREP_NUP_NROW}` or 2 if the variable is not set.
    ncol
        The number of columns on each output sheet.
        The default is `${REPREP_NUP_NCOL}` or 2 if the variable is not set.
    margin
        The margin in mm between the pages on each sheet. (Also used as sheet margin.)
        The default is `${REPREP_NUP_MARGIN}` or 10.0 if the variable is not set.
    page_format
        The output page format
        The default is `${REPREP_NUP_PAGE_FORMAT}` or A4-L if the variable is not set.
    optional
        When `True`, the step is only executed when needed by other steps.
    block
        When `True`, the step will always remain pending.

    Returns
    -------
    step_info
        Holds relevant information of the step, useful for defining follow-up steps.
    """
    command = "python -m stepup.reprep.nup_pdf ${inp} ${out}"
    if nrow is not None:
        command += f" -r {nrow}"
    if ncol is not None:
        command += f" -c {ncol}"
    if margin is not None:
        command += f" -m {margin}"
    if page_format is not None:
        command += f" -p {page_format}"
    return step(command, inp=path_src, out=path_dst, optional=optional, block=block)


def raster_pdf(
    path_inp: str,
    out: str,
    *,
    resolution: int | None = None,
    quality: int | None = None,
    optional: bool = False,
    block: bool = False,
) -> StepInfo:
    """Turn each page of a PDF into a rendered JPEG bitmap contained in a new PDF.

    Parameters
    ----------
    path_inp
        The input PDF file.
    out
        None, output directory or path. See `make_path_out`.
    resolution
        The resolution of the bitmap in dots per inch (pdi).
        The default value is taken from `${REPREP_RASTER_RESOLUTION}` or 100 if the variable is not
        set.
    quality
        The JPEG quality of the bitmap.
        The default value is taken from `${REPREP_RASTER_QUALITY}` or 50 if the variable is not set.
    optional
        When `True`, the step is only executed when needed by other steps.
    block
        When `True`, the step will always remain pending.

    Returns
    -------
    step_info
        Holds relevant information of the step, useful for defining follow-up steps.
    """
    command = "python -m stepup.reprep.raster_pdf ${inp} ${out}"
    if resolution is not None:
        command += f" -r {resolution}"
    if quality is not None:
        command += f" -q {quality}"
    path_out = make_path_out(path_inp, out, ".pdf")
    return step(command, inp=path_inp, out=path_out, optional=optional, block=block)


def render(
    path_template: str,
    paths_variables: list[str],
    out: str,
    *,
    mode: str = "auto",
    optional: bool = False,
    block: bool = False,
) -> StepInfo:
    """Render the template with Jinja2.

    Parameters
    ----------
    path_template
        The source file to use as a template.
    paths_variables
        A list of Python files with variable definitions, at least one.
    out
        An output directory or file.
    mode
        The format of the Jinja placeholders.
        The default (auto) selects either plain or latex based on the extension of the template.
        The plain format is the default Jinja style with curly brackets: {{ }} etc.
        The latex style replaces curly brackets by angle brackets: << >> etc.
    optional
        When `True`, the step is only executed when needed by other steps.
    block
        When `True`, the step will always remain pending.

    Returns
    -------
    step_info
        Holds relevant information of the step, useful for defining follow-up steps.
    """
    if mode not in ["auto", "plain", "latex"]:
        raise ValueError(f"Unsupported mode {mode!r}. Must be one of 'auto', 'plain', 'latex'")
    if len(paths_variables) == 0:
        raise ValueError("At least one file with variable definitions needed.")
    path_out = make_path_out(path_template, out, None)
    command = "python -m stepup.reprep.render ${inp} ${out}"
    if mode != "auto":
        command += f" --mode {mode}"
    return step(
        command,
        inp=[path_template, *paths_variables],
        out=path_out,
        optional=optional,
        block=block,
    )


def sync_zenodo(path_config: str, *, block: bool = False) -> StepInfo:
    """Synchronize data with an draft dataset on Zenodo.

    Parameters
    ----------
    path_config
        The YAML configuration file for the Zenodo upload.
    block
        When `True`, the step will always remain pending.

    Returns
    -------
    step_info
        Holds relevant information of the step, useful for defining follow-up steps.
    """
    return step("python -m stepup.reprep.sync_zenodo ${inp}", inp=path_config, block=block)


def unplot(
    path_svg: str, out: str | None = None, *, optional: bool = False, block: bool = False
) -> StepInfo:
    """Convert a plot back to data.

    Parameters
    ----------
    path_svg
        The SVG file with paths to be converted back.
    out
        An output directory or file.

    optional
        When `True`, the step is only executed when needed by other steps.
    block
        When `True`, the step will always remain pending.

    Returns
    -------
    step_info
        Holds relevant information of the step, useful for defining follow-up steps.
    """
    path_out = make_path_out(path_svg, out, ".json")
    command = "python -m stepup.reprep.unplot ${inp} ${out}"
    return step(command, inp=path_svg, out=path_out, optional=optional, block=block)


def zip_inventory(
    path_inventory: str, path_zip: str, *, optional: bool = False, block: bool = False
) -> StepInfo:
    """Create a ZIP file with all files listed in a `inventory.txt` file + check digests before zip.

    Parameters
    ----------
    path_inventory
        A file created with the `make_inventory` API or with the command-line script
        `reprep-make-inventory`.
    path_zip
        The output ZIP file
    optional
        When `True`, the step is only executed when needed by other steps.
    block
        When `True`, the step will always remain pending.

    Returns
    -------
    step_info
        Holds relevant information of the step, useful for defining follow-up steps.
    """
    return step(
        "python -m stepup.reprep.zip_inventory ${inp} ${out}",
        inp=path_inventory,
        out=path_zip,
        optional=optional,
        block=block,
    )



#!/usr/bin/env python
from stepup.core.api import static
from stepup.reprep.api import latex_flat

static("article_structured.tex", "part1.tex", "part2.tex")
latex_flat("article_structured.tex", "article.tex")



# StepUp RepRep is the StepUp extension for Reproducible Reporting.
# Copyright (C) 2024 Toon Verstraelen
#
# This file is part of StepUp RepRep.
#
# StepUp RepRep is free software;  you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# StepUp RepRep is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Unit tests for stepup.reprep.render."""

from stepup.reprep.render import load_module_file


def test_load_module_file(path_tmp):
    path_foo = path_tmp / "foo.py"
    with open(path_foo, "w") as fh:
        print("a = 10", file=fh)
    foo = load_module_file(path_foo)
    assert foo.a == 10


def test_load_module_file2(path_tmp):
    path_foo = path_tmp / "foo.py"
    with open(path_foo, "w") as fh:
        print("a = 10", file=fh)
    path_bar = path_tmp / "bar.py"
    with open(path_bar, "w") as fh:
        print("from foo import a", file=fh)
    bar = load_module_file(path_bar)
    assert bar.a == 10



# StepUp RepRep is the StepUp extension for Reproducible Reporting.
# Copyright (C) 2024 Toon Verstraelen
#
# This file is part of StepUp RepRep.
#
# StepUp RepRep is free software;  you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# StepUp RepRep is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Synchronization of local datasets with drafst on Zenodo.

This script synchronizes your local version of a dataset
with the (draft of) this dataset on Zenodo.
This simplifies your interaction with Zenodo to the following steps:

1. Prepare a personal token
2. Write metadata in a configuration YAML file.
3. Include this script in your workflow to automatically sync data.
4. Publish the dataset through the Zenodo web interface.

After publication, only metadata can be updated, not the actual files.
If you need to upload different files, increment the version number.
This will result in a new draft that you can publish when it is final.

This script offers a simplified interface to Zenodo.
It does not try to mirror all of Zenodo's REST API features.
Zenodo has subtle differences in APIs for drafts and submitted records,
which are normalized in the `ZenodoWrapper` class to become identical.
"""

import argparse
import hashlib
import json
import sys
from datetime import date
from typing import Any

import attrs
import cattrs
import markdown
import requests
import semver
import yaml
from path import Path

from stepup.core.api import amend


class RESTError(Exception):
    """Raised when a REST API call is not successful."""


class ZenodoError(Exception):
    """Raised in case of an error in the logic of the main script."""


@attrs.define
class RESTWrapper:
    """Wrapper for REST APIs that use JSON input (mostly) and output (always)."""

    endpoint: str = attrs.field()
    params: dict[str, str] = attrs.field()
    verbose: bool = attrs.field(default=False)

    def request(self, method: str, loc: str, **kwargs) -> Any:
        """Send a HTTP request and deserialize the response as JSON.

        Parameters
        ----------
        method
            The HTTP method: GET, POST, PUT or DELETE.
        loc
            The address to be appended after the endpoint.
        kwargs
            Keyword arguments to pass on to the `requests.request` function.

        Returns
        -------
        response_data
            Deserialized JSON response data.
        """
        url = f"{self.endpoint}/{loc}"
        if self.verbose:
            print(f"{method} {url}")
        res = requests.request(method, url, params=self.params, **kwargs)
        data = None if len(res.text) == 0 else res.json()
        if not res.ok:
            raise RESTError(
                f"Failed {method} {url}: {res.status_code}\n" + json.dumps(data, indent=2)
            )
        if self.verbose:
            print(json.dumps(data, indent=2))
        return data

    def get(self, loc: str, **kwargs):
        """Create a GET HTTP requests. See `request` method for details."""
        return self.request("GET", loc, **kwargs)

    def post(self, loc: str, **kwargs):
        """Create a POST HTTP requests. See `request` method for details."""
        return self.request("POST", loc, **kwargs)

    def put(self, loc: str, **kwargs):
        """Create a PUT HTTP requests. See `request` method for details."""
        return self.request("PUT", loc, **kwargs)

    def delete(self, loc: str, **kwargs):
        """Create a DELETE HTTP requests. See `request` method for details."""
        return self.request("DELETE", loc, **kwargs)


@attrs.define
class Creator:
    """A Zenodo creator"""

    name: str = attrs.field(converter=str.strip)
    """Formatted as `last, first`."""

    affiliation: str = attrs.field(converter=str.strip)


@attrs.define
class Metadata:
    """A subset of the Zenodo metadata."""

    title: str = attrs.field(validator=attrs.validators.min_len(1))
    version: str = attrs.field(converter=str.strip)
    license: str = attrs.field(converter=lambda s: s.strip().lower())
    upload_type: str = attrs.field(
        validator=attrs.validators.in_(
            [
                "publication",
                "poster",
                "presentation",
                "dataset",
                "image",
                "video",
                "software",
                "lesson",
                "physicalobject",
                "other",
            ]
        )
    )
    creators: list[Creator] = attrs.field()
    description: str | None = attrs.field(default=None)


@attrs.define
class File:
    """A subset of the Zenodo file."""

    version_id: str = attrs.field()
    name: str = attrs.field()
    size: int = attrs.field()
    checksum: str = attrs.field()
    links: dict[str, str] = attrs.field()


@attrs.define
class Record:
    """A subset of the Zenodo record API.

    For a full definition, see:
    https://github.com/zenodo/zenodo/tree/master/zenodo/modules/deposit/jsonschemas/deposits/records
    """

    record_id: int | None = attrs.field()
    metadata: Metadata = attrs.field()
    links: dict[str, str] = attrs.field()
    files: list[File] = attrs.field()
    state: str = attrs.field()
    submitted: bool = attrs.field()


@attrs.define
class ZenodoWrapper:
    """Python interface to a subset of the Zenodo API."""

    token: str = attrs.field()
    endpoint: str = attrs.field(default="https://sandbox.zenodo.org/api")
    verbose: bool = attrs.field(default=False)
    rest: RESTWrapper = attrs.field(init=False)

    @rest.default
    def _default_rest(self):
        """Set the default RESP Wrapper."""
        return RESTWrapper(self.endpoint, {"access_token": self.token}, verbose=self.verbose)

    def create_new_record(self, metadata: Metadata) -> Record:
        """Create a new record on Zenodo, which remains in draft until it is published manually."""
        data = {"metadata": self._unstructure_metadata(metadata)}
        res = self.rest.post("deposit/depositions", json=data)
        self._normalize_record(res)
        return cattrs.structure(res, Record)

    def get_record(self, record_id: int) -> Record:
        """Get an (un)published record with given id.

        Note that Zenodo has different APIs for submitted and draft records.
        This method tries both and normalizes the record so that the caller
        does not have to deal with those differences.
        """
        try:
            # If the dataset is in draft
            res = self.rest.get(f"deposit/depositions/{record_id}")
        except RESTError:
            # If the dataset is published
            res = self.rest.get(f"records/{record_id}")
        self._normalize_record(res)
        return cattrs.structure(res, Record)

    def update_record(self, record_id: int, metadata: Metadata) -> Record:
        """Update the metadata of a record.

        This is applicable draft records and published records in edit mode.
        """
        data = {"metadata": self._unstructure_metadata(metadata)}
        data["metadata"]["publication_date"] = date.today().isoformat()
        res = self.rest.put(f"deposit/depositions/{record_id}", json=data)
        self._normalize_record(res)
        return cattrs.structure(res, Record)

    def edit_record(self, record_id: int):
        """Put a published record into edit mode."""
        self.rest.post(f"deposit/depositions/{record_id}/actions/edit")

    def publish_record(self, record_id: int):
        """Publish are draft record or a record in edit mode."""
        self.rest.post(f"deposit/depositions/{record_id}/actions/publish")

    def upload_file(self, bucket: str, path: str, name: str) -> File:
        """Upload a file. The buckect must belong to a record in draft mode."""
        with open(path, "rb") as fh:
            res = self.rest.put(f"{bucket}/{name}", data=fh)
        # Normalize file info before structuring.
        res["checksum"] = res["checksum"][4:]
        res["name"] = res["key"]
        self._normalize_links(res["links"])
        return cattrs.structure(res, File)

    def delete_file(self, record_id: int, file_id: str):
        """Delete a file. The file must belong to a record in draft mode."""
        self.rest.delete(f"deposit/depositions/{record_id}/files/{file_id}")

    def create_new_version(self, record_id: int) -> Record:
        """Create a new version of a published record. The result is a draft record."""
        res = self.rest.post(f"deposit/depositions/{record_id}/actions/newversion")
        self._normalize_record(res)
        return cattrs.structure(res, Record)

    def _unstructure_metadata(self, metadata: Metadata) -> dict[str]:
        """Turn a Metadata instance into a JSON-able dictionary."""
        result = cattrs.unstructure(metadata)
        self._normalize_metadata(result)
        return result

    def _normalize_record(self, record: dict[str]):
        """Normalize a record received from Zenodo."""
        if "id" in record and "record_id" not in record:
            record["record_id"] = record["id"]
        self._normalize_links(record.get("links"))
        self._normalize_metadata(record.get("metadata"))
        for file in record.get("files", []):
            self._normalize_file(file)

    def _normalize_file(self, file: dict[str]):
        """Normalize a file received from Zenodo."""
        file["version_id"] = file["id"]
        if "checksum" in file and file["checksum"].startswith("md5:"):
            file["checksum"] = file["checksum"][4:]
        if "size" not in file and "filesize" in file:
            file["size"] = file["filesize"]
        if "name" not in file:
            if "filename" in file:
                file["name"] = file["filename"]
            elif "key" in file:
                file["name"] = file["key"]
        self._normalize_links(file["links"])

    def _normalize_links(self, links: dict[str, str]):
        """Normalize the links received from Zenodo: remove non-API urls and remove endpoint."""
        for key, url in list(links.items()):
            if not url.startswith(self.endpoint):
                del links[key]
            else:
                links[key] = url[len(self.endpoint) + 1 :]

    @staticmethod
    def _normalize_metadata(metadata: dict[str]):
        """Normalize metadata receivd from Zenodo or taken from the YAML config."""
        if metadata["description"] is None:
            del metadata["description"]
        if "upload_type" not in metadata and "resource_type" in metadata:
            metadata["upload_type"] = metadata["resource_type"]["type"]
        if isinstance(metadata["license"], dict) and "id" in metadata["license"]:
            metadata["license"] = metadata["license"]["id"]
        if "version" not in metadata:
            metadata["version"] = "TODO"


@attrs.define
class Config:
    """Configuration datastructure.

    An object of this class is created from data loaded from a local YAML config file.
    It provides a more convenient way to access the configuration data.
    """

    path_versions: str = attrs.field()
    endpoint: str = attrs.field()
    path_token: str = attrs.field()
    metadata: Metadata = attrs.field()
    path_readme: str | None = attrs.field(default=None)
    paths: list[str] = attrs.field(factory=list)


def main(argv: list[str] | None = None):
    """Main program."""
    args = parse_args(argv)
    with open(args.config) as fh:
        data = yaml.safe_load(fh)
        if not isinstance(data.get("metadata").get("version"), str):
            raise TypeError(
                "The version in the YAML config file is missing or is not a string. "
                "Enclose the versionin quotes to make it a string."
            )
        config = cattrs.structure(data, Config)
    update_online(config, args.verbose)


def parse_args(argv: list[str] | None = None) -> argparse.Namespace:
    """Parse command-line arguments."""
    parser = argparse.ArgumentParser(
        prog="reprep-sync-zenodo", description="Sync a draft dataset on Zenodo."
    )
    parser.add_argument("config", help="Configuration YAML file.")
    parser.add_argument(
        "--verbose",
        "-v",
        default=False,
        action="store_true",
        help="Show details of communication with Zenodo endpoint.",
    )
    return parser.parse_args(argv)


def load_version(path_versions: str, version: str) -> int:
    """Load the record_id for a given version from a versions.json file."""
    path_versions = Path(path_versions)
    if not path_versions.is_file():
        return None
    with open(path_versions) as fh:
        versions = json.load(fh)
    if len(versions) == 0:
        return None
    record_id = versions.get(version)
    if record_id is not None:
        return record_id
    return max(versions.values())


def dump_version(path_versions: str, version: str, record_id: int):
    """Dump the record_id for a given version into a versions.json file."""
    path_versions = Path(path_versions)
    if not path_versions.is_file():
        versions = {}
    else:
        with open(path_versions) as fh:
            versions = json.load(fh)
    versions[version] = record_id
    # Serialize before opening file,
    # so no information is lost when JSON fails.
    serialized = json.dumps(versions, indent=2)
    with open(path_versions, "w") as fh:
        fh.write(serialized)
        fh.write("\n")


def update_online(config: Config, verbose: bool):
    """Make the online data set up to date with the local information."""
    # Amend inputs when this script is called in a StepUp workflow.
    paths_inp = list(config.paths)
    if config.path_readme is not None:
        paths_inp.append(config.path_readme)
    if not amend(inp=paths_inp):
        print("No amend")
        return

    # If present, convert README Markdown file to HTML
    if config.path_readme is not None:
        if not config.path_readme.endswith(".md"):
            raise ValueError("The README Markdown file must end with the .md extension.")
        md_ctx = markdown.Markdown(extensions=["fenced_code"])
        with open(config.path_readme) as fh:
            config.metadata.description = md_ctx.convert(fh.read())

    # Run sanity check on paths: duplicate filenames not allowed.
    paths = {}
    for path in config.paths:
        path = Path(path)
        paths[path.name] = path
    if len(paths) != len(config.paths):
        raise ZenodoError(
            "Zenodo does not support directory layouts. Files must have a different names."
        )

    # (Try to) get the token.
    path_token = Path(config.path_token).expand()
    if not path_token.is_file():
        return
    with open(path_token) as fh:
        zenodo = ZenodoWrapper(fh.read().strip(), config.endpoint, verbose=verbose)

    # Interact with Zenodo.
    record_id = load_version(config.path_versions, config.metadata.version)
    if record_id is None:
        # New record, when getting started with a dataset.
        record = _create_new(zenodo, config.metadata, paths, config.metadata.version)
        dump_version(config.path_versions, config.metadata.version, record.record_id)
    else:
        # When a dataset exists, the actions depend on the current status of the record.
        record = zenodo.get_record(record_id)
        if record.submitted:
            cmpver = semver.compare(config.metadata.version, record.metadata.version)
            if cmpver == 0:
                _check_record_md5(record, paths, config.metadata.version)
                if record.metadata != config.metadata:
                    _republish_metadata(zenodo, record_id, config.metadata)
            elif cmpver > 0:
                record = _create_new_version(zenodo, record_id, config.metadata)
                dump_version(config.path_versions, config.metadata.version, record.record_id)
                _refresh_files(zenodo, record, paths, config.metadata.version)
            else:
                raise ZenodoError(
                    f"The online version ({record.metadata.version}) is newer "
                    f"than the local one ({config.metadata.version})."
                )
        else:
            changed = _refresh_files(zenodo, record, paths, config.metadata.version)
            if changed or record.metadata != config.metadata:
                zenodo.update_record(record_id, config.metadata)


def _create_new(
    zenodo: ZenodoWrapper, metadata: Metadata, paths: dict[str, Path], version: str
) -> Record:
    """Create a new record on Zenodo."""
    record = zenodo.create_new_record(metadata)
    for name, path in paths.items():
        print(f"Uploading {path}")
        file = zenodo.upload_file(record.links["bucket"], path, name)
        if not _match_md5(path, file.checksum):
            raise ZenodoError(f"MD5 Checksum mismatch for {path} ({version}, new)")
    return record


def _match_md5(path: str, checksum: str) -> bool:
    """Compute the MD5 sum of a file and compare to the given checksum."""
    with open(path, "rb") as fh:
        return checksum == hashlib.file_digest(fh, hashlib.md5).hexdigest()


def _check_record_md5(record: Record, paths: dict[str, Path], version: str):
    """Sanity check of MD5 hashes received from Zenodo"""
    for file in record.files:
        if file.name not in paths:
            raise ZenodoError(
                f"File {file.name} exists online but not locally ({version}, published)"
            )
        path = paths[file.name]
        if not _match_md5(path, file.checksum):
            raise ValueError(f"MD5 Checksum mismatch for {path} ({version}, published)")
    online_names = {file.name for file in record.files}
    for name in paths:
        if name not in online_names:
            raise OSError(f"File {name} exists locally but not online ({version}, published)")


def _republish_metadata(zenodo: ZenodoWrapper, record_id: int, metadata: Metadata):
    """Put a record in edit mode, update the metadata and publish again."""
    print(f"Editing metadata and publishing same version ({metadata.version}) again.")
    zenodo.edit_record(record_id)
    zenodo.update_record(record_id, metadata)
    zenodo.publish_record(record_id)


def _create_new_version(zenodo: ZenodoWrapper, record_id: int, metadata: Metadata) -> Record:
    """Create a new version of the dataset and refresh the metadata."""
    print(f"Creating a new version ({metadata.version})")
    record = zenodo.create_new_version(record_id)
    return zenodo.update_record(record.record_id, metadata)


def _refresh_files(
    zenodo: ZenodoWrapper, record: Record, paths: dict[str, Path], version: str
) -> bool:
    """Refresh the online files.

    This function only uploads files that do not exit yet online or have changed locally.
    Files removed from the local YAML config will also be removed online.
    """
    changed = False
    for file in record.files:
        if file.name not in paths:
            print(f"Deleting {file.name} ({version}, draft)")
            zenodo.delete_file(record.record_id, file.version_id)
            changed = True
        else:
            path = paths[file.name]
            if not _match_md5(path, file.checksum):
                print(f"Replacing {path} ({version}, draft)")
                zenodo.delete_file(record.record_id, file.version_id)
                file = zenodo.upload_file(record.links["bucket"], path, file.name)
                if not _match_md5(path, file.checksum):
                    raise ZenodoError(f"MD5 Checksum mismatch for {path} ({version}, draft)")
                changed = True
    online_names = {file.name for file in record.files}
    for name, path in paths.items():
        if name not in online_names:
            print(f"Uploading {path} ({version}, draft)")
            file = zenodo.upload_file(record.links["bucket"], path, name)
            if not _match_md5(path, file.checksum):
                raise ZenodoError(f"MD5 Checksum mismatch for {path} ({version}, draft)")
            changed = True
    return changed


if __name__ == "__main__":
    main(sys.argv[1:])



# StepUp RepRep is the StepUp extension for Reproducible Reporting.
# Copyright (C) 2024 Toon Verstraelen
#
# This file is part of StepUp RepRep.
#
# StepUp RepRep is free software;  you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# StepUp RepRep is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""RepRep Wrapper for LaTeX."""

import argparse
import subprocess
import sys

from path import Path

from stepup.core.api import amend, getenv
from stepup.core.hash import compute_file_digest
from stepup.reprep.make_inventory import write_inventory

from .bibtex_log import parse_bibtex_log
from .latex_deps import scan_latex_deps
from .latex_log import ErrorInfo, parse_latex_log


def main(argv: list[str] | None = None):
    """Main program."""
    args = parse_args(argv)

    workdir, fn_tex = Path(args.path_tex).splitpath()
    workdir = workdir.normpath()
    if not fn_tex.endswith(".tex"):
        raise ValueError("The LaTeX source must have extension .tex")
    stem = fn_tex[:-4]
    path_aux = workdir / f"{stem}.aux"

    # Remove existing outputs from a previous run, which could potentially
    # conflict with the new tex source files. In 99% of the cases, this is
    # not a problem, but sometimes LaTeX chokes on remnants in old outputs.
    exts_to_remove = ["log", "aux", "blg", "fls", "out", "toc", "nlo", "synctex"]
    if args.run_bibtex:
        exts_to_remove.append("bbl")
    for ext in exts_to_remove:
        (workdir / f"{stem}.{ext}").remove_p()

    # Detect additional inputs
    implicit, bib = scan_latex_deps(fn_tex)

    # Get LaTeX executable
    if args.latex is None:
        args.latex = getenv("REPREP_LATEX", "pdflatex")

    aux_digest_hist = []
    if len(bib) == 0:
        if not amend(inp=implicit):
            sys.exit(3)
        inventory_files = list(implicit)
    elif args.run_bibtex:
        # Get other executables and files
        if args.bibtex is None:
            args.bibtex = getenv("REPREP_BIBTEX", "bibtex")
        if args.bibsane is None:
            args.bibsane = getenv("REPREP_BIBSANE", "bibsane")
        paths_config = []
        if args.bibsane_config is None:
            args.bibsane_config = getenv("REPREP_BIBSANE_CONFIG", "bibsane.yaml", is_path=True)
            paths_config.append(args.bibsane_config)

        if not amend(
            inp=implicit + bib + paths_config,
            out=[f"{stem}.bbl"],
        ):
            sys.exit(3)
        inventory_files = [*implicit, *bib, f"{stem}.bbl"]

        # LaTeX
        cp = subprocess.run(
            [
                args.latex,
                "-interaction=errorstopmode",
                "-draftmode",
                stem,
            ],
            stdin=subprocess.DEVNULL,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
            check=False,
            cwd=workdir,
        )
        if cp.returncode != 0:
            path_log = workdir / f"{stem}.log"
            error_info = parse_latex_log(path_log)
            error_info.print(path_log)
            sys.exit(1)

        aux_digest_hist.append(compute_file_digest(path_aux))

        # BibTeX
        cp = subprocess.run(
            [args.bibtex, stem],
            stdin=subprocess.DEVNULL,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
            check=False,
            cwd=workdir,
        )
        if cp.returncode != 0:
            path_blg = workdir / f"{stem}.blg"
            error_info = parse_bibtex_log(path_blg)
            error_info.print(path_blg)
            sys.exit(1)

        # BibSane
        cp = subprocess.run(
            [args.bibsane, f"{stem}.aux", f"--config={args.bibsane_config}"],
            cwd=workdir,
            text=True,
            stdin=subprocess.DEVNULL,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            check=False,
        )
        if cp.returncode != 0:
            error_info = ErrorInfo("BibSane", src=f"{workdir}/{stem}.aux")
            error_info.print()
            sys.stdout.write(cp.stdout)
            sys.exit(1)
    else:
        if not amend(inp=[*implicit, f"{stem}.bbl"]):
            sys.exit(3)
        inventory_files = [*implicit, f"{stem}.bbl"]

    for _ in range(args.maxrep):
        # LaTeX
        cp = subprocess.run(
            [args.latex, "-interaction=errorstopmode", stem],
            stdin=subprocess.DEVNULL,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
            check=False,
            cwd=workdir,
        )
        path_log = workdir / f"{stem}.log"
        error_info = parse_latex_log(path_log)
        if cp.returncode != 0:
            error_info.print(path_log)
            sys.exit(1)
        aux_digest_hist.append(compute_file_digest(path_aux))
        if len(aux_digest_hist) > 1 and aux_digest_hist[-1] == aux_digest_hist[-2]:
            break
    else:
        print(
            f"\033[1;31;40mAux file did not converge in {args.maxrep} iterations!\033[0;0m",
            file=sys.stderr,
        )
        print(path_aux, file=sys.stderr)
        for digest in aux_digest_hist:
            print(digest.hex(), file=sys.stderr)
        sys.exit(1)

    inventory_files.extend([f"{stem}.tex", f"{stem}.aux", f"{stem}.pdf"])
    path_inventory = f"{stem}-inventory.txt"
    write_inventory(path_inventory, inventory_files)

    vol_paths = []
    for path in Path(".").glob(f"{stem}.*"):
        path = path.normpath()
        if not (path in inventory_files or path == path_inventory):
            vol_paths.append(path)
    amend(vol=vol_paths)


def parse_args(argv: list[str] | None = None) -> argparse.Namespace:
    """Parse command-line arguments."""
    parser = argparse.ArgumentParser(
        prog="reprep-latex",
        description="Compile a LaTeX document and print essence from log file.",
    )
    parser.add_argument("path_tex", help="The main LaTeX source file.")
    parser.add_argument(
        "-m",
        "--maxrep",
        default=5,
        type=int,
        help="The maximum number of LaTeX recompilations (not including the one for BibTeX).",
    )
    parser.add_argument(
        "--latex",
        help="The LaTeX executable. "
        "The default is ${REPREP_LATEX} or pdflatex if the variable is not defined.",
    )
    parser.add_argument(
        "--run-bibtex",
        dest="run_bibtex",
        default=False,
        action="store_true",
        help="Run bibtex and bibsane.",
    )
    parser.add_argument(
        "--bibtex",
        help="The BibTeX executable. "
        "The default is ${REPREP_BIBTEX} or bibtex if the variable is not defined.",
    )
    parser.add_argument(
        "--bibsane",
        help="The BibSane executable. "
        "The default is ${REPREP_BIBSANE} or bibsane if the variable is not defined.",
    )
    parser.add_argument(
        "--bibsane-config",
        help="The BibSane configuration file. The default is ${REPREP_BIBSANE_CONFIG} or "
        "bibsane.yaml if the variables is not defined.",
    )
    return parser.parse_args(argv)


if __name__ == "__main__":
    main(sys.argv[1:])



#!/usr/bin/env python
from stepup.core.api import static, step
from stepup.reprep.api import convert_markdown

static("macros.tex")
for i in range(20):
    fn_random = f"random_{i:02d}.md"
    step(f"./write.py {i} > {fn_random}", out=fn_random)
    convert_markdown(fn_random, katex=True, path_macro="macros.tex")



#!/usr/bin/env python

from stepup.core.api import static, step
from stepup.reprep.api import latex

static("paper.tex")
static("smile.pdf")
latex("paper.tex")
step("echo 'Hi there!' > generated.tex", out=["generated.tex"])



# StepUp RepRep is the StepUp extension for Reproducible Reporting.
# Copyright (C) 2024 Toon Verstraelen
#
# This file is part of StepUp RepRep.
#
# StepUp RepRep is free software;  you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# StepUp RepRep is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Rendering of template files with Jinja2.

Parameters for the template can be defined in a "variables" Python file,
from which all variables that are strings, integers or floats are imported.
"""

import argparse
import contextlib
import importlib.util
import sys
from types import ModuleType

import jinja2
from path import Path


def main(argv: list[str] | None = None):
    """Main program."""
    args = parse_args(argv)
    if args.mode == "plain":
        latex = False
    elif args.mode == "latex":
        latex = True
    elif args.mode == "auto":
        latex = args.path_out.endswith(".tex")
    else:
        raise ValueError(f"mode not supported: {args.mode}")
    dir_out = Path(args.path_out).parent.absolute()
    variables = load_variables(args.paths_variables, dir_out)
    result = render(args.path_in, variables, latex)
    with open(args.path_out, "w") as fh:
        fh.write(result)
    return 0


def parse_args(argv: list[str] | None = None) -> argparse.Namespace:
    """Parse command-line arguments."""
    parser = argparse.ArgumentParser(prog="reprep-render", description="Render a file with Jinja2.")
    parser.add_argument("path_in", help="The input file")
    parser.add_argument("paths_variables", nargs="+", help="Python files defining variables")
    parser.add_argument("path_out", help="The output file")
    parser.add_argument(
        "--mode",
        choices=["auto", "plain", "latex"],
        help="The delimiter style to use",
        default="auto",
    )
    return parser.parse_args(argv)


def load_variables(paths_variables: list[str], dir_out: str) -> dict[str, str]:
    """Load user-defined variable from Python files.

    Parameters
    ----------
    paths_variables
        paths of Python files containing variable definitions.
        They are loaded in the given order, so later variable definitions may overrule earlier ones.
    dir_out
        This is used to translate paths defined the variables files to relative paths with respect
        to the parent of the output of the rendering task.

    Returns
    -------
    variables
        A dictionary with variables.
    """
    variables = {}
    for path_py in paths_variables:
        path_py = Path(path_py)
        dir_py = path_py.parent.normpath()
        fn_py = path_py.name
        with contextlib.chdir(dir_py):
            module = load_module_file(fn_py, "variables")
            current = vars(module)
        for name, value in current.items():
            if isinstance(value, (str | int | float | None)):
                if isinstance(value, Path):
                    value = value.relpath(dir_out)
                variables[name] = value
    return variables


def load_module_file(path_py: str, name: str = "pythonscript") -> ModuleType:
    """Load a Python module from a given path."""
    parent = Path(path_py).parent
    sys.path.append(parent)
    try:
        spec = importlib.util.spec_from_file_location(f"<{name}>", str(path_py))
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
    finally:
        sys.path.remove(parent)
    return module


def render(
    path_template: str,
    variables: dict[str, str],
    latex: bool = False,
    *,
    str_in: str | None = None,
) -> str:
    """The template is processed with jinja and returned after filling in variables.

    Parameters
    ----------
    path_template
        The filename of the template to load, may be a mock
    variables
        A dictionary of variables to substitute into the template.
    latex
        When True, the angle-version of the template codes is used, e.g. `<%` etc.
    str_in
        The template string.
        When given path_templates is not loaded and only used for error messages.

    Returns
    -------
    str_out
        A string with the result.
    """
    # Customize Jinja 2 environment
    env_kwargs = {
        "keep_trailing_newline": True,
        "trim_blocks": True,
        "undefined": jinja2.StrictUndefined,
        "autoescape": False,
    }
    if latex:
        env_kwargs.update(
            {
                "block_start_string": "<%",
                "block_end_string": "%>",
                "variable_start_string": "<<",
                "variable_end_string": ">>",
                "comment_start_string": "<#",
                "comment_end_string": "#>",
                "line_statement_prefix": "%==",
            }
        )
    env = jinja2.Environment(**env_kwargs)

    # Load template and use it
    if str_in is None:
        with open(path_template) as f:
            str_in = f.read()
    template = env.from_string(str_in)
    template.filename = path_template
    return template.render(**variables)


if __name__ == "__main__":
    main(sys.argv[1:])



#!/usr/bin/env python
import random
import sys

TEXT_SNIPPETS = """
Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor
incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud
exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat.
""".split()


EQ_SNIPPETS = r"""
x y \sin(\theta) \hat{H} \vec{B} \int_0^\infty f(x)\,dx
\{i^2\}_{i=1}^N e^{2it\pi\omega} \tan(x)
""".split()


def main():
    """Write a randomized markdown file."""
    parts = []
    random.seed(int(sys.argv[1]))
    for _ in range(100):
        parts.append(random.choice(TEXT_SNIPPETS))
        parts.append(f"$`{random.choice(EQ_SNIPPETS)}`$")
    print(" ".join(parts))


if __name__ == "__main__":
    main()



# StepUp RepRep is the StepUp extension for Reproducible Reporting.
# Copyright (C) 2024 Toon Verstraelen
#
# This file is part of StepUp RepRep.
#
# StepUp RepRep is free software;  you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# StepUp RepRep is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Unit tests for inventory files."""

import contextlib
import os

import pytest
from path import Path

from stepup.reprep.check_inventory import main as check_main
from stepup.reprep.make_inventory import main as make_main
from stepup.reprep.make_inventory import parse_inventory_def
from stepup.reprep.zip_inventory import main as zip_main

BASIC_INVENTORY = """\
              4 -rw-r--r-- b25ac67550816e84f2a5689115534c8b1d72081f14f6bc4395e4af41c1dfa83163d5433a\
e05e4cba54e04364d38f9084ec9c3e41d0fbe45da3b738939b1aec11 a.txt
              4 -rw-r--r-- 10901f8827ee64f82d48afd9fc4a9bf9befaea8cf0b00a4bf16b1533cae88c04f7b467b4\
c59a690462e898e75a0def6c0168b3aef7d01600944f811b4b75da35 b.txt
"""


def test_basic(path_tmp):
    with contextlib.chdir(path_tmp):
        with open("a.txt", "w") as fh:
            print("aaa", file=fh)
        with open("b.txt", "w") as fh:
            print("bbb", file=fh)
        with open("inventory.def", "w") as fh:
            print("include *.txt", file=fh)
        make_main(["-i", "inventory.def"])
        with open("inventory.txt") as fh:
            assert fh.read() == BASIC_INVENTORY
        make_main(["-o", "inventory2.txt", "a.txt", "b.txt"])
        with open("inventory2.txt") as fh:
            assert fh.read() == BASIC_INVENTORY
        check_main(["inventory.txt"])


SYMLINK_INVENTORY = """\
              5 -rw-r--r-- ef15eaf92d5e335345a3e1d977bc7d8797c3d275717cc1b10af79c93cda01aeb2a0c59bc\
02e2bdf9380fd1b54eb9e1669026930ccc24bd49748e65f9a6b2ee68 dest.txt
                lrwxrwxrwx 35d3c7e91999a8d52f496686b07466d03b7e4d85f2c9a85e2b391df1047cb57a6d1c39b8\
4ace64e72267ec116e3376d47fb4512327688eeecb88cf9883e0738a link.txt
"""


def test_symbolic_link(path_tmp):
    with contextlib.chdir(path_tmp):
        with open("dest.txt", "w") as fh:
            fh.write("Hello")
        Path("link.txt").symlink_to("dest.txt")
        make_main(["-o", "inventory.txt", "dest.txt", "link.txt"])
        with open("inventory.txt") as fh:
            assert fh.read() == SYMLINK_INVENTORY
        check_main(["inventory.txt"])


def test_directory(path_tmp):
    with pytest.raises(OSError):
        make_main(["-o", path_tmp / "inventory.txt", path_tmp])


SYMLINK_DIR_INVENTORY = """\
                lrwxrwxrwx a645e8e6808560219f3710f0d46e4ad2cef0b0745a9ee2605ff8446e3edb7bfe600de66d\
d018a886303fceb7be6c614021b7bc706ca03052b39d43899129e20b link
"""


def test_symbolic_link_directory(path_tmp):
    path_sub = path_tmp / "sub"
    path_sub.mkdir()
    with contextlib.chdir(path_tmp):
        Path("link").symlink_to("sub")
        make_main(["-o", "inventory.txt", "link"])
        with open("inventory.txt") as fh:
            assert fh.read() == SYMLINK_DIR_INVENTORY
        check_main(["inventory.txt"])


def test_git():
    with contextlib.chdir("tests/cases/check_hrefs_md"):
        paths = parse_inventory_def(["include-git\n"])
    assert paths == {
        "main.sh",
        "expected_graph.txt",
        "plan.py",
        "test.md",
        "expected_stdout.txt",
        "README.md",
        "check_hrefs.yaml",
        ".gitignore",
    }


def test_zip_inventory(path_tmp):
    with contextlib.chdir(path_tmp):
        Path("sub").mkdir()
        with open("sub/dest.txt", "w") as fh:
            fh.write("Hello")
        Path("link.txt").symlink_to("sub/dest.txt")
        Path("link").symlink_to("sub")
        make_main(["-o", "inventory2.txt", "sub/dest.txt", "link.txt", "link"])
        check_main(["inventory2.txt"])
        zip_main(["inventory2.txt", "archive.zip"])
        Path("sub/dest.txt").remove()
        Path("sub/").rmdir()
        Path("link.txt").remove()
        Path("link").remove()
        Path("inventory2.txt").move("inventory1.txt")
        os.system("unzip archive.zip")
        check_main(["inventory1.txt"])
        check_main(["inventory2.txt"])



# StepUp RepRep is the StepUp extension for Reproducible Reporting.
# Copyright (C) 2024 Toon Verstraelen
#
# This file is part of StepUp RepRep.
#
# StepUp RepRep is free software;  you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# StepUp RepRep is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Utilities shared by several unit test modules."""

import contextlib


@contextlib.contextmanager
def local_file(contents, filename, tmpdir):
    """Change to a temporary directory and create a file with given contents."""
    with contextlib.chdir(tmpdir):
        with open(filename, "w") as fh:
            fh.write(contents)
        yield
